// SPDX-FileCopyrightText: 2022 Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: AGPL-3.0-or-later

mod config;
mod constants;

use {
    self::config::Config,
    anyhow::anyhow,
    base64::write::EncoderWriter as Base64Encoder,
    clap::{arg, command, ArgMatches, Command},
    reqwest::Url,
    std::io::Write,
};

fn parse_args() -> ArgMatches {
    let server_arg = arg!(-s --server <server> "NextCloud server to connect to").required(false);

    command!()
        .propagate_version(true)
        .subcommand_required(true)
        .arg_required_else_help(true)
        .subcommand(
            Command::new("init")
                .about("Authenticate against the provided NextCloud server")
                .arg(arg!(<server> "NextCloud server to connect to")),
        )
        .subcommand(
            Command::new("import")
                .about("Import the given URLs into NextCloud's cookbook")
                .arg(server_arg)
                .arg(arg!(<url> ... "One or more URLs each pointing to page with a recipe to import in NextCloud")),
        )
        .get_matches()
}

#[tokio::main(flavor = "multi_thread")]
async fn main() -> anyhow::Result<()> {
    let args = parse_args();
    let mut configuration = Config::new();

    match args.subcommand() {
        Some(("init", sub_matches)) => {
            let server = sub_matches
                .get_one::<String>("server")
                .expect("Mandatory parameter <server>");
            tokio::task::block_in_place(move || -> anyhow::Result<()> {
                configuration
                    .credentials
                    .add(Url::parse(server)?)
                    .expect("Unable to authenticate to NextCloud instance");
                Ok(())
            })?;
        }
        Some(("import", sub_matches)) => {
            let server_name = sub_matches
                .get_one::<String>("server")
                .unwrap_or_else(|| {
                    let servers = &configuration.credentials.servers;
                    match servers.len() {
                        0 => panic!("No NextCloud server set up yet, use '{} init' first", env!("CARGO_BIN_NAME")),
                        1 => servers.iter().next().unwrap().0,
                        _ => panic!("More than one NextCloud server set up, use the '--server' option to specify which one to use. Known servers: {:#?}", servers.keys().collect::<Vec<_>>()),
                    }
                });

            let api_client = ApiClient::new(&server_name, &configuration)?;
            for url in sub_matches
                .get_many::<String>("url")
                .expect("At least one url is required")
            {
                let response = api_client
                    .client
                    .post(api_client.base_url.join("apps/cookbook/import")?)
                    .json(&serde_json::json!({
                        "url": url,
                    }))
                    .send()
                    .await?;
                println!("{:#?}", response); // TODO
            }
        }
        _ => unreachable!("Exhausted list of subcommands and subcommand_required prevents `None`"),
    };

    Ok(())
}

struct ApiClient {
    pub base_url: Url,
    pub client: reqwest::Client,
}

impl ApiClient {
    pub fn new(server_name: &str, configuration: &Config) -> anyhow::Result<Self> {
        let server = configuration
            .credentials
            .servers
            .get(server_name)
            .ok_or_else(|| {
                anyhow!(
                    "Unknown server {}. Did you use '{} init' first? Known servers: {:#?}",
                    server_name,
                    env!("CARGO_BIN_NAME"),
                    configuration.credentials.servers.keys().collect::<Vec<_>>()
                )
            })?;

        use reqwest::header;
        let mut default_headers = header::HeaderMap::new();
        let mut auth_header = b"Basic ".to_vec();
        {
            let mut encoder = Base64Encoder::new(&mut auth_header, base64::STANDARD);
            write!(encoder, "{}:{}", server.login_name, server.password).unwrap();
        }
        let mut auth_header = header::HeaderValue::from_bytes(&auth_header)?;
        auth_header.set_sensitive(true);
        default_headers.insert(header::AUTHORIZATION, auth_header);

        let client = reqwest::Client::builder()
            .user_agent(constants::USER_AGENT)
            .default_headers(default_headers)
            .build()?;

        Ok(ApiClient {
            base_url: Url::parse(&server.url)?,
            client: client,
        })
    }
}
